// Map.h

#ifndef MAP_H_ROGUELIKE_1
#define MAP_H_ROGUELIKE_1

#include "Tile.h"
#include "Terrain.h"
#include "Creature.h"

struct Offset
{
  Offset(short newX = 0, short newY = 0)
    : x(newX), y(newY) {}

  bool operator==(Offset const & right) const
  {
    return x == right.x && y == right.y;
  }

  Offset operator+(Offset const & right) const
  {
    return Offset(x+right.x, y+right.y);
  }
  Offset & operator+=(Offset const & right)
  {
    x += right.x;
    y += right.y;
    return *this;
  }
  short x;
  short y;
};

std::ostream & operator<<(std::ostream & stream, Offset const & right)
{
  stream << "(" << right.x << ", " << right.y << ")";
  return stream;
}

struct MapCell
{
public:
  MapCell()
    : facingCount(0)
    , view(TILE_UNKNOWN)
  {}

  MapCell(MapCell const & right)
    : facingCount(right.facingCount)
    , view(right.view)
    , place(right.place)
  {
    if (right.person.get() != NULL)
    {
      person.reset(new Creature(* right.person));
    }
  }

  MapCell & operator=(MapCell const & right)
  {
    MapCell temp(right);
    std::swap(view, temp.view);

    std::auto_ptr<Creature> tempCreature(temp.person);
    temp.person = person;
    person = tempCreature;

    std::swap(facingCount, temp.facingCount);
    std::swap(place, temp.place);
    return *this;
  }
  char facingCount;
  Tile view;
  std::auto_ptr<Creature> person;
  Terrain place;
};

class Map
{
public:
  Map(int newSizeX = 1, int newSizeY = 1)
  {
    reset(newSizeX, newSizeY);
  }

  void reset(int newSizeX, int newSizeY)
  {
    assert(newSizeX > 0);
    assert(newSizeY > 0);
    cells.resize(static_cast<size_t>(newSizeX*newSizeY));
    defaultCell.view = TILE_UNKNOWN;
    defaultCell.place.blocked = true;
    defaultCell.place.view = TILE_UNKNOWN;
    sizeX = newSizeX;
    sizeY = newSizeY;
  }

  MapCell & operator[](Offset const & pos)
  {
//    assert(pos.x >= 0 && pos.x < sizeX);
//    assert(pos.y >= 0 && pos.y < sizeY);
    if (pos.x < 0 || pos.x >= sizeX || pos.y < 0 || pos.y >= sizeY)
    {
      return defaultCell;
    }
    else
    {
      return cells[static_cast<size_t>(pos.x + pos.y*sizeX)];
    }
  }

  int getSizeX(void)
  {
    return sizeX;
  }

  int getSizeY(void)
  {
    return sizeY;
  }

private:
  std::vector<MapCell> cells;
  MapCell defaultCell;
  int sizeX;
  int sizeY;
};

Tile calculateTile(Offset const & pos)
{
  Tile result = TILE_UNKNOWN;
  MapCell & current = (*g::currentMap)[pos];
  result = current.place.view;
  if (current.person.get() != NULL)
  {
    result = current.person->view;
  }
  if (current.facingCount > 0)
  {
    result = static_cast<Tile>(result | TILE_FACING);
  }
  return result;
}

class ShowToPlayer
{
public:
  bool operator()(Offset const & /*source*/, Offset const & pos)
  {
//    (*g::currentMap)[pos].playerVisible = true;
    (*g::currentMap)[pos].view = calculateTile(pos);
    return false;
  }
};

class HideFromPlayer
{
public:
  bool operator()(Offset const & /*source*/, Offset const & pos)
  {
//    (*g::currentMap)[pos].playerVisible = false;
    (*g::currentMap)[pos].view = static_cast<Tile>(calculateTile(pos)
                                                   | TILE_LAST_SEEN);
    return false;
  }
};

#endif
