// main.cc

#include "lib.h"
#include "interface.h"
#include "Map.h"
#include "Los.h"

using namespace std;

namespace g
{
  Map * currentMap;
  ofstream logFile;

  Offset playerPos;
  Offset playerFacing;
  Los<ShowToPlayer> showLos;
  Los<HideFromPlayer> hideLos;
}

enum MoveAction
{
  MOVE_PLAYER = 1,
  FACE_PLAYER = 2,
  MOVE_AND_FACE_PLAYER = 3
};

void movePlayer(Offset move, MoveAction action)
{
  Offset newPos = g::playerPos + move;
  auto_ptr<Creature> player = (*g::currentMap)[g::playerPos].person;
  (*g::currentMap)[g::playerPos + g::playerFacing].facingCount--;
    g::hideLos.calculateFov(g::playerPos, HideFromPlayer());
//  g::hideLos.calculateFovQuadrant(g::playerPos, HideFromPlayer(),
//                                  g::playerFacing);
  if ((action & MOVE_PLAYER) && ! (*g::currentMap)[newPos].place.blocked)
  {
    g::playerPos = newPos;
  }
  if (action & FACE_PLAYER)
  {
    g::playerFacing = move;
  }

  (*g::currentMap)[g::playerPos].person = player;
  (*g::currentMap)[g::playerPos + g::playerFacing].facingCount++;
  g::showLos.calculateFov(g::playerPos, ShowToPlayer());
//  g::showLos.calculateFovQuadrant(g::playerPos, ShowToPlayer(),
//                                  g::playerFacing);
}

void showMap(Offset center)
{
  Offset corner(center.x - (MapViewer::getWidth() - 1)/2,
                center.y - (MapViewer::getHeight() - 1)/2);
  for (int x = 0; x < MapViewer::getWidth(); ++x)
  {
    for (int y = 0; y < MapViewer::getHeight(); ++y)
    {
      MapCell & current = (*g::currentMap)[corner + Offset(x, y)];
//      if (current.playerVisible)
//      {
        MapViewer::setTile(x, y, current.view);
//      }
//      else
//      {
//        MapViewer::setTile(x, y, TILE_UNKNOWN);
//      }
    }
  }
  MapViewer::display();
}

void init(string mapName)
{
  g::logFile.open("err.txt", ios::out | ios::trunc);

  Interface::init();
  MapViewer::init();
  KeyboardReader::init();

  Creature player;
  player.view = TILE_PLAYER;

  Terrain wall;
  wall.blocked = true;
  wall.view = TILE_WALL;

  Terrain floor;
  floor.blocked = false;
  floor.view = TILE_FLOOR;

  ifstream mapFile(mapName.c_str(), ios::in);
  if (!mapFile)
  {
    cerr << "Couldn't open map file: " << mapName << endl;
    exit(1);
  }
  string line;
  getline(mapFile, line);
  for (int y = 0; y < g::currentMap->getSizeY(); ++y)
  {
    for (int x = 0; x < g::currentMap->getSizeX(); ++x)
    {
      Offset pos(x, y);
      if (mapFile && x < static_cast<int>(line.size()))
      {
        if (line[x] == '#')
        {
          (*g::currentMap)[pos].place = wall;
        }
        else if (line[x] == '@')
        {
          (*g::currentMap)[pos].place = floor;
          (*g::currentMap)[pos].person.reset(new Creature(player));
          g::playerPos = pos;
          g::playerFacing = Offset(1,0);
        }
        else
        {
          (*g::currentMap)[pos].place = floor;
        }
      }
      else
      {
        (*g::currentMap)[pos].place = wall;
      }
      (*g::currentMap)[pos].view = TILE_UNKNOWN;
    }
    if (mapFile)
    {
      getline(mapFile, line);
    }
  }
  (*g::currentMap)[g::playerPos + g::playerFacing].facingCount = 1;
//  g::showLos.calculateFovQuadrant(g::playerPos, ShowToPlayer(),
//                                  g::playerFacing);
  g::showLos.calculateFov(g::playerPos, ShowToPlayer());
  showMap(g::playerPos);
}

void cleanup(void)
{
  KeyboardReader::cleanup();
  MapViewer::cleanup();
  Interface::cleanup();
}

void mainLoop(void)
{
  bool done = false;
  while (!done)
  {
    Key current = KeyboardReader::getNextKey();
    Offset move;
    MoveAction action = MOVE_AND_FACE_PLAYER;
    switch (current & KEY_SIMPLE_MASK)
    {
    case KEY_NUMPAD_1:
      move.x = -1;
      move.y = 1;
      break;
    case KEY_NUMPAD_2:
      move.x = 0;
      move.y = 1;
      break;
    case KEY_NUMPAD_3:
      move.x = 1;
      move.y = 1;
      break;
    case KEY_NUMPAD_4:
      move.x = -1;
      move.y = 0;
      break;
    case KEY_NUMPAD_6:
      move.x = 1;
      move.y = 0;
      break;
    case KEY_NUMPAD_7:
      move.x = -1;
      move.y = -1;
      break;
    case KEY_NUMPAD_8:
      move.x = 0;
      move.y = -1;
      break;
    case KEY_NUMPAD_9:
      move.x = 1;
      move.y = -1;
      break;
    case KEY_NUMPAD_0:
      done = true;
      // No break. Fall through.
    case KEY_NUMPAD_5:
    case KEY_UNKNOWN:
    default:
      move.x = 0;
      move.y = 0;
      break;
    }
    if (move.x != 0 || move.y != 0)
    {
      if (current & KEY_SHIFT)
      {
        action = MOVE_PLAYER;
      }
      if (current & KEY_CONTROL)
      {
        action = FACE_PLAYER;
      }
      movePlayer(move, action);
    }
    showMap(g::playerPos);
  }
}

int main(int argc, char * argv[])
{
  string mapName = "map.txt";
  if (argc == 2)
  {
    mapName = argv[1];
  }
  Map theMap(100, 100);
  g::currentMap = &theMap;
  init(mapName);
  mainLoop();
  cleanup();
  return 0;
}
